"""
API wrapper for Outline VPN
"""

import typing
from dataclasses import dataclass

import requests
from urllib3 import PoolManager

UNABLE_TO_GET_METRICS_ERROR = "Unable to get metrics"


@dataclass
class OutlineKey:
    """
    Describes a key in the Outline server
    """

    key_id: str
    name: str
    password: str
    port: int
    method: str
    access_url: str
    used_bytes: int
    data_limit: typing.Optional[int]

    def __init__(self, response: dict, metrics: dict = None):
        self.key_id = response.get("id")
        self.name = response.get("name")
        self.password = response.get("password")
        self.port = response.get("port")
        self.method = response.get("method")
        self.access_url = response.get("accessUrl")
        self.used_bytes = (
            metrics.get("bytesTransferredByUserId").get(response.get("id"))
            if metrics
            else 0
        )
        self.data_limit = response.get("dataLimit", {}).get("bytes")


class OutlineServerErrorException(Exception):
    pass


class OutlineLibraryException(Exception):
    pass


class _FingerprintAdapter(requests.adapters.HTTPAdapter):
    """
    This adapter injected into the requests session will check that the
    fingerprint for the certificate matches for every request
    """

    def __init__(self, fingerprint=None, **kwargs):
        self.fingerprint = str(fingerprint)
        super(_FingerprintAdapter, self).__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            assert_fingerprint=self.fingerprint,
        )


class OutlineVPN:
    """
    An Outline VPN connection
    """

    def __init__(self, api_url: str, cert_sha256: str, cert_path: str = None):
        self.api_url = api_url

        if cert_sha256:
            session = requests.Session()
            session.mount("https://", _FingerprintAdapter(cert_sha256))
            session.verify = cert_path if cert_path else False
            self.session = session
        else:
            raise OutlineLibraryException(
                "No certificate SHA256 provided. Running without certificate is no longer supported."
            )

    def get_keys(self, timeout: int = None) -> list[OutlineKey]:
        """Get all keys in the outline server"""
        response = self.session.get(
            f"{self.api_url}/access-keys/", timeout=timeout
        )
        if response.status_code == 200 and "accessKeys" in response.json():
            response_metrics = self.session.get(
                f"{self.api_url}/metrics/transfer"
            )
            if (
                response_metrics.status_code >= 400
                or "bytesTransferredByUserId" not in response_metrics.json()
            ):
                raise OutlineServerErrorException(UNABLE_TO_GET_METRICS_ERROR)

            response_json = response.json()
            result: list[OutlineKey] = []
            for key in response_json.get("accessKeys"):
                result.append(OutlineKey(key, response_metrics.json()))
            return result
        raise OutlineServerErrorException("Unable to retrieve keys")

    def get_key(self, key_id: str, timeout: int = None) -> OutlineKey:
        response = self.session.get(
            f"{self.api_url}/access-keys/{key_id}", timeout=timeout
        )
        if response.status_code == 200:
            key = response.json()

            response_metrics = self.session.get(
                f"{self.api_url}/metrics/transfer", timeout=timeout
            )
            if (
                response_metrics.status_code >= 400
                or "bytesTransferredByUserId" not in response_metrics.json()
            ):
                raise OutlineServerErrorException(UNABLE_TO_GET_METRICS_ERROR)

            return OutlineKey(key, response_metrics.json())
        else:
            raise OutlineServerErrorException("Unable to get key")

    def create_key(
        self,
        key_id: str = None,
        name: str = None,
        method: str = None,
        password: str = None,
        data_limit: int = None,
        port: int = None,
        timeout: int = None,
    ) -> OutlineKey:
        """Create a new key"""

        payload = {}
        if name:
            payload["name"] = name
        if method:
            payload["method"] = method
        if password:
            payload["password"] = password
        if data_limit:
            payload["limit"] = {"bytes": data_limit}
        if port:
            payload["port"] = port
        if key_id:
            payload["id"] = key_id
            response = self.session.put(
                f"{self.api_url}/access-keys/{key_id}",
                verify=False,
                json=payload,
                timeout=timeout,
            )
        else:
            response = self.session.post(
                f"{self.api_url}/access-keys",
                verify=False,
                json=payload,
                timeout=timeout,
            )

        if response.status_code == 201:
            key = response.json()
            outline_key = OutlineKey(key)
            return outline_key

        raise OutlineServerErrorException(f"Unable to create key. {response.text}")

    def delete_key(self, key_id: str, timeout: int = None) -> bool:
        """Delete a key"""
        response = self.session.delete(
            f"{self.api_url}/access-keys/{key_id}", timeout=timeout
        )
        return response.status_code == 204

    def rename_key(self, key_id: str, name: str, timeout: int = None):
        """Rename a key"""
        files = {
            "name": (None, name),
        }

        response = self.session.put(
            f"{self.api_url}/access-keys/{key_id}/name",
            files=files,
            verify=False,
            timeout=timeout,
        )
        return response.status_code == 204

    def add_data_limit(
        self, key_id: str, limit_bytes: int, timeout: int = None
    ) -> bool:
        """Set data limit for a key (in bytes)"""
        data = {"limit": {"bytes": limit_bytes}}

        response = self.session.put(
            f"{self.api_url}/access-keys/{key_id}/data-limit",
            json=data,
            verify=False,
            timeout=timeout,
        )
        return response.status_code == 204

    def delete_data_limit(self, key_id: str, timeout: int = None) -> bool:
        """Removes data limit for a key"""
        response = self.session.delete(
            f"{self.api_url}/access-keys/{key_id}/data-limit",
            verify=False,
            timeout=timeout,
        )
        return response.status_code == 204

    def get_transferred_data(self, timeout: int = None):
        """Gets how much data all keys have used
        {
            "bytesTransferredByUserId": {
                "1":1008040941,
                "2":5958113497,
                "3":752221577
            }
        }"""
        response = self.session.get(
            f"{self.api_url}/metrics/transfer", timeout=timeout
        )
        if (
            response.status_code >= 400
            or "bytesTransferredByUserId" not in response.json()
        ):
            raise OutlineServerErrorException(UNABLE_TO_GET_METRICS_ERROR)
        return response.json()

    def get_server_information(self, timeout: int = None):
        """Get information about the server
        {
            "name":"My Server",
            "serverId":"7fda0079-5317-4e5a-bb41-5a431dddae21",
            "metricsEnabled":true,
            "createdTimestampMs":1536613192052,
            "version":"1.0.0",
            "accessKeyDataLimit":{"bytes":8589934592},
            "portForNewAccessKeys":1234,
            "hostnameForAccessKeys":"example.com"
        }
        """
        response = self.session.get(
            f"{self.api_url}/server", timeout=timeout
        )
        if response.status_code != 200:
            raise OutlineServerErrorException(
                "Unable to get information about the server"
            )
        return response.json()

    def set_server_name(self, name: str, timeout: int = None) -> bool:
        """Renames the server"""
        data = {"name": name}
        response = self.session.put(
            f"{self.api_url}/name", json=data, timeout=timeout
        )
        return response.status_code == 204

    def set_hostname(self, hostname: str, timeout: int = None) -> bool:
        """Changes the hostname for access keys.
        Must be a valid hostname or IP address."""
        data = {"hostname": hostname}
        response = self.session.put(
            f"{self.api_url}/server/hostname-for-access-keys",
            verify=False,
            json=data,
            timeout=timeout,
        )
        return response.status_code == 204

    def get_metrics_status(self, timeout: int = None) -> bool:
        """Returns whether metrics is being shared"""
        response = self.session.get(
            f"{self.api_url}/metrics/enabled", timeout=timeout
        )
        return response.json().get("metricsEnabled")

    def set_metrics_status(self, status: bool, timeout: int = None) -> bool:
        """Enables or disables sharing of metrics"""
        data = {"metricsEnabled": status}
        response = self.session.put(
            f"{self.api_url}/metrics/enabled", json=data, timeout=timeout
        )
        return response.status_code == 204

    def set_port_new_for_access_keys(self, port: int, timeout: int = None) -> bool:
        """Changes the default port for newly created access keys.
        This can be a port already used for access keys."""
        data = {"port": port}
        response = self.session.put(
            f"{self.api_url}/server/port-for-new-access-keys",
            verify=False,
            json=data,
            timeout=timeout,
        )
        if response.status_code == 400:
            raise OutlineServerErrorException(
                "The requested port wasn't an integer from 1 through 65535, or the request had no port parameter."
            )
        elif response.status_code == 409:
            raise OutlineServerErrorException(
                "The requested port was already in use by another service."
            )
        return response.status_code == 204

    def set_data_limit_for_all_keys(
        self, limit_bytes: int, timeout: int = None
    ) -> bool:
        """Sets a data transfer limit for all access keys."""
        data = {"limit": {"bytes": limit_bytes}}
        response = self.session.put(
            f"{self.api_url}/server/access-key-data-limit",
            verify=False,
            json=data,
            timeout=timeout,
        )
        return response.status_code == 204

    def delete_data_limit_for_all_keys(self, timeout: int = None) -> bool:
        """Removes the access key data limit, lifting data transfer restrictions on all access keys."""
        response = self.session.delete(
            f"{self.api_url}/server/access-key-data-limit",
            verify=False,
            timeout=timeout,
        )
        return response.status_code == 204



"""
Integration tests for the API wrapper
"""

import json
import random
import re

import pytest

from outline_vpn.outline_vpn import OutlineVPN, OutlineLibraryException


@pytest.fixture
def client() -> OutlineVPN:
    """This generates a client from the credentials provided in the environment"""
    install_log = open("outline-install.log", "r").read()
    json_text = re.findall("({[^}]+})", install_log)[0]
    api_data = json.loads(json_text)
    api_url = re.sub("https://[^:]+:", "https://127.0.0.1:", api_data.get("apiUrl"))

    client = OutlineVPN(api_url=api_url, cert_sha256=api_data.get("certSha256"), cert_path="./shadowbox-selfsigned.crt")

    return client


def test_no_cert_sha256_raises_exception():
    """Test that the client raises an exception if the cert sha256 is not provided"""
    with pytest.raises(OutlineLibraryException):
        OutlineVPN(api_url="https://aaa", cert_sha256="")


def test_get_keys(client: OutlineVPN):  # pylint: disable=W0621
    """Test for the get keys method"""
    assert len(client.get_keys()) >= 1


def test_crud_key(client: OutlineVPN):  # pylint: disable=W0621
    """Test creating/updating the name/deleting a key"""
    new_key = client.create_key()
    assert new_key is not None
    assert int(new_key.key_id) > 0

    read_key = client.get_key(new_key.key_id)
    assert read_key is not None
    assert read_key.key_id == new_key.key_id

    named_key = client.create_key(name="Test Key")
    assert named_key.name == "Test Key"

    assert client.rename_key(new_key.key_id, "a_name")

    assert client.delete_key(new_key.key_id)


def test_create_key_with_attributes(client: OutlineVPN):
    """Test creating a key with attributes"""
    key = client.create_key(
        name="Another test key",
        data_limit=1024 * 1024 * 20,
        method="aes-192-gcm",
        password="test",
        port=4545,
    )
    assert key.name == "Another test key"
    assert key.method == "aes-192-gcm"
    assert key.password == "test"
    assert key.data_limit == 1024 * 1024 * 20
    assert key.port == 4545
    assert client.delete_key(key.key_id)


def test_create_key_with_attributes_and_id(client: OutlineVPN):
    """Test creating a key with attributes and id"""
    key_id = f"test-id-{random.random()}"
    key = client.create_key(
        key_id=key_id,
        name="Yet another test key",
        data_limit=1024 * 1024 * 20,
        method="aes-192-gcm",
        password="test",
        port=2323,
    )
    assert key.key_id == key_id
    assert key.name == "Yet another test key"
    assert key.method == "aes-192-gcm"
    assert key.password == "test"
    assert key.data_limit == 1024 * 1024 * 20
    assert key.port == 2323
    assert client.delete_key(key.key_id)


def test_limits(client: OutlineVPN):  # pylint: disable=W0621
    """Test setting, retrieving and removing custom limits"""
    new_key = client.create_key()
    assert new_key is not None
    assert int(new_key.key_id) > 0

    new_limit = 1024 * 1024 * 20

    assert client.add_data_limit(new_key.key_id, new_limit)

    keys = client.get_keys()
    for key in keys:
        if key.key_id == new_key.key_id:
            assert key.data_limit == new_limit

    assert client.delete_data_limit(new_key.key_id)


def test_server_methods(client: OutlineVPN):
    server_info = client.get_server_information()
    assert server_info is not None
    new_server_name = "Test Server name"
    assert client.set_server_name(new_server_name)
    new_hostname = "example.com"
    assert client.set_hostname(new_hostname)
    new_port_for_access_keys = 11233
    assert client.set_port_new_for_access_keys(new_port_for_access_keys)
    updated_server_info = client.get_server_information()
    assert updated_server_info.get("name") == new_server_name
    assert updated_server_info.get("hostnameForAccessKeys") == new_hostname
    assert updated_server_info.get("portForNewAccessKeys") == new_port_for_access_keys

    assert client.set_server_name(server_info.get("name"))
    assert client.set_hostname(server_info.get("hostnameForAccessKeys"))
    assert client.set_port_new_for_access_keys(server_info.get("portForNewAccessKeys"))


def test_metrics_status(client: OutlineVPN):
    metrics_status = client.get_metrics_status()
    assert client.set_metrics_status(not metrics_status)
    assert client.get_metrics_status() != metrics_status
    client.set_metrics_status(metrics_status)


def test_data_limit_for_all_keys(client: OutlineVPN):
    assert client.set_data_limit_for_all_keys(1024 * 1024 * 20)
    assert client.delete_data_limit_for_all_keys()


def test_get_transferred_data(client: OutlineVPN):
    """Call the method and assert it responds something"""
    data = client.get_transferred_data()
    assert data is not None
    assert "bytesTransferredByUserId" in data
